from typing import Any, Union, Callable, List, Dict, Tuple
from contextlib import suppress

from .common import CHECK_TYPES
from .typesystem import TypeSystem, PythonBasic
from .dispatch import MultiDispatch

dp = MultiDispatch(PythonBasic())


def _isinstance(a, b):
    try:
        return isinstance(a, b)
    except TypeError as e:
        raise TypeError(f"Bad arguments to isinstance: {a}, {b}") from e

_orig_issubclass = issubclass
def _issubclass(a, b):
    try:
        return _orig_issubclass(a, b)
    except TypeError as e:
        raise TypeError(f"Bad arguments to issubclass: {a}, {b}") from e


class SubclassDispatch(PythonBasic):
    isinstance = issubclass

dp_type = MultiDispatch(SubclassDispatch())

class RuntypeError(TypeError):
    pass

class TypeMistmatchError(RuntypeError):
    pass

class TupleLengthError(TypeMistmatchError):
    pass


def switch_subclass(t, d):
    for k, v in d.items():
        if _issubclass(t, k):
            return v


def ensure_isa_seq(obj, t):
    ensure_isa(obj, list)
    for item in obj:
        ensure_isa(item, t.__args__)

def ensure_isa_tuple(obj, t):
    ensure_isa(obj, tuple)
    if len(obj) != len(t.__args__):
        raise TupleLengthError(obj, t.__args__)
    for item, type_ in zip(obj, t.__args__):
        ensure_isa(item, type_)

def ensure_isa_dict(obj, t):
    ensure_isa(obj, dict)
    kt, vt = t.__args__
    for k, v in obj.items():
        ensure_isa(k, kt)
        ensure_isa(v, vt)


def ensure_isa(obj, t):
    if type(t) is tuple:
        if not any(isa(obj, opt) for opt in t):
            raise TypeMistmatchError(obj, t)
        return
    try:
        t.__origin__
    except AttributeError:
        if t is Any or t == (Any,):
            return
        if not _isinstance(obj, t):
            raise TypeMistmatchError(obj, t)
    else:
        if t.__origin__ is Union:
            ensure_isa(obj, t.__args__)    # Send as tuple
        elif t is Callable:
            if not callable(obj):
                raise TypeMistmatchError(obj, callable)
        else:
            assert t.__origin__, t
            f = switch_subclass(t.__origin__, {
                (list, set): ensure_isa_seq,
                tuple: ensure_isa_tuple,
                dict: ensure_isa_dict,
            })
            f(obj, t)



def isa(obj, t):
    try:
        ensure_isa(obj, t)
        return True
    except TypeMistmatchError as e:
        return False


def assert_isa(obj, t):
    if CHECK_TYPES:
        try:
            ensure_isa(obj, t)
        except TypeMistmatchError as e:
            item_value, item_type = e.args
            msg = f"Expected value of type {t}, instead got {obj!r}"
            if item_value is not obj:
                msg += f'\n\n    Failed on item: {item_value}, expected type {item_type}'
            raise TypeError(msg)


def canonize_type(t):
    "Turns List -> list, Dict -> dict, etc."
    try:
        return {
            Any: object,
            List: list,
            Dict: dict,
            Tuple: tuple,
            List[Any]: list,
            Dict[Any, Any]: dict,
            Tuple[Any]: tuple,
        }[t]
    except KeyError:
        with suppress(AttributeError):
            if t.__origin__ is Union:
                return t.__args__
        return t

def issubclass(t1, t2):
    if t2 is Any:
        return True

    t1 = canonize_type(t1)

    if isinstance(t1, tuple):
        return all(issubclass(t, t2) for t in t1)
    elif isinstance(t2, tuple):
        return any(issubclass(t1, t) for t in t2)

    if hasattr(t2, '__origin__'):
        t2 = canonize_type(t2)
        return t1 == t2    # TODO add some clever logic here
    elif hasattr(t1, '__origin__'):
        return issubclass(t1.__origin__, t2)    # XXX more complicated than that?

    return _issubclass(t1, t2)



class PythonTyping(TypeSystem):
    isinstance = staticmethod(isa)
    issubclass = staticmethod(issubclass)
    canonize_type = staticmethod(canonize_type)
    get_type = type
    default_type = object


from copy import copy
import dataclasses

from .common import CHECK_TYPES
from .isa import ensure_isa as default_ensure_isa, TypeMistmatchError

def _post_init(self, ensure_isa):
    for name, field in getattr(self, '__dataclass_fields__', {}).items():
        value = getattr(self, name)
        try:
            ensure_isa(value, field.type)
        except TypeMistmatchError as e:
            item_value, item_type = e.args
            msg = f"[{type(self).__name__}] Attribute '{name}' expected value of type {field.type}, instead got {value!r}"
            if item_value is not value:
                msg += f'\n\n    Failed on item: {item_value}, expected type {item_type}'
            raise TypeError(msg)


def _setattr(self, name, value, ensure_isa):
    try:
        field = self.__dataclass_fields__[name]
    except (KeyError, AttributeError):
        pass
    else:
        ensure_isa(value, field.type)


def replace(self, **kwargs):
    """Returns a new instance, with the given attibutes and values overwriting the existing ones.

    Useful for making copies with small updates.

    Examples:
        >>> @dataclass
        ... class A:
        ...     a: int
        ...     b: int
        >>> A(1, 2).replace(a=-2)
        A(a=-2, b=2)

        >>> some_instance.replace() == copy(some_instance)   # Equivalent operations
        True
    """
    return dataclasses.replace(self, **kwargs)

def __iter__(self):
    "Yields a list of tuples [(name, value), ...]"
    return ((name, getattr(self, name)) for name in self.__dataclass_fields__)

def aslist(self):
    """Returns a list of values

    Equivalent to: list(dict(this).values())
    """
    return [getattr(self, name) for name in self.__dataclass_fields__]

def astuple(self):
    """Returns a tuple of values

    Equivalent to: tuple(dict(this).values())
    """
    return tuple(getattr(self, name) for name in self.__dataclass_fields__)


def _set_if_not_exists(cls, d):
    for attr, value in d.items():
        try:
            getattr(cls, attr)
        except AttributeError:
            setattr(cls, attr, value)


def _process_class(cls, ensure_isa, check_types, **kw):
    if check_types:
        c = copy(cls)

        orig_post_init = getattr(cls, '__post_init__', None)
        def __post_init__(self):
            _post_init(self, ensure_isa=ensure_isa)
            if orig_post_init is not None:
                orig_post_init(self)
        c.__post_init__ = __post_init__

        if not kw['frozen']:
            orig_set_attr = getattr(cls, '__setattr__')
            def __setattr__(self, name, value):
                _setattr(self, name, value, ensure_isa=ensure_isa)
                orig_set_attr(self, name, value)
            c.__setattr__ = __setattr__
    else:
        c = cls

    _set_if_not_exists(c, {
        'replace': replace,
        'aslist': aslist,
        'astuple': astuple,
        '__iter__': __iter__,
    })
    return dataclasses.dataclass(c, **kw)


def dataclass(cls=None, *, ensure_isa=default_ensure_isa, check_types=CHECK_TYPES, init=True, repr=True, eq=True, order=False, unsafe_hash=False, frozen=True):
    """runtype.dataclass is a drop-in replacement, that adds functionality on top of Python's built-in dataclass.

    * Adds run-time type validation
    * Adds convenience methods:
        * replace - create a new instance, with updated attributes
        * aslist - returns the dataclass values as a list
        * astuple - returns the dataclass values as a tuple
        * dict(this) - returns a dict of the dataclass attributes and values
    * Frozen by default. Can be disabled (not recommended)

    Note: Changes to an existing instance (i.e. by setting attributes) are not validated!
          Use `frozen=False` at your own risk.

    Example:
        >>> @dataclass
        >>> class Point:
        ...     x: int
        ...     y: int

        >>> p = Point(2, 3)
        >>> p
        Point(x=2, y=3)
        >>> dict(p)         # Maintains order
        {'x': 2, 'y': 3}

        >>> p.replace(x=30)  # New instance
        Point(x=30, y=3)
    """

    def wrap(cls):
        return _process_class(cls, ensure_isa, check_types, init=init, repr=repr, eq=eq, order=order, unsafe_hash=unsafe_hash, frozen=frozen)

    # See if we're being called as @dataclass or @dataclass().
    if cls is None:
        # We're called with parens.
        return wrap

    # We're called as @dataclass without parens.
    return wrap(cls)





class TypeSystem:
    def isinstance(self, obj, t):
        return self.issubclass(self.get_type(obj), t)

    def issubclass(self, t1, t2):
        raise NotImplementedError()

    def canonize_type(self, t):
        return t

    def get_type(self, obj):
        raise NotImplementedError()

    default_type = NotImplemented



class PythonBasic(TypeSystem):
    isinstance = isinstance
    issubclass = issubclass
    get_type = type
    default_type = object



from .isa import isa, issubclass, assert_isa, PythonTyping, TypeMistmatchError
from .dispatch import MultiDispatch, DispatchError
from .dataclass import dataclass


def Dispatch(typesystem=PythonTyping()):
    return MultiDispatch(typesystem)



__version__ = "0.1.7"



from collections import defaultdict
import inspect
from functools import wraps

class DispatchError(Exception):
    pass

class MultiDispatch:
    """Create a decorator attached to a dispatch group,
    that when applied to a function, enables multiple-dispatch for it.

    Example:
        >>> from runtype import Dispatch
        >>> dp = Dispatch()

        >>> @dp
        ... def add1(i: Optional[int]):
        ...     return i + 1

        >>> @dp
        ... def add1(s: Optional[str]):
        ...     return s + "1"

        >>> @dp
        ... def add1(a):  # Any, which is the least-specific
        ...     return (a, 1)

        >>> add1(1)
        2

        >>> add1("1")
        11

        >>> add1(1.0)
        (1.0, 1)


    """
    def __init__(self, typesystem):
        self.roots = defaultdict(TypeTree)
        self.typesystem = typesystem

    def __call__(self, f):
        root = self.roots[f.__name__]
        root.name = f.__name__
        root.typesystem = self.typesystem

        root.define_function(f)

        @wraps(f)
        def dispatched_f(*args, **kw):
            f = root.find_function_cached(args)
            return f(*args, **kw)

        return dispatched_f


class TypeTree:
    def __init__(self):
        self.root = TypeNode()
        self._cache = {}
        self.name = None
        self.typesystem = None

    def get_args_simple_signature(self, args):
        get_type = self.typesystem.get_type
        return tuple(get_type(a) for a in args)

    def find_function(self, args):
        nodes = [self.root]
        for a in args:
            nodes = [n for node in nodes for n in node.follow_arg(a, self.typesystem)]

        funcs = [node.func for node in nodes if node.func]

        if len(funcs) == 0:
            raise DispatchError(f"Function '{self.name}' not found for signature {self.get_args_simple_signature(args)}")
        elif len(funcs) > 1:
            f, _sig = self.choose_most_specific_function(*funcs)
        else:
            (f, _sig) ,= funcs
        return f


    def find_function_cached(self, args):
        "Memoized version of find_function"
        sig = self.get_args_simple_signature(args)
        try:
            return self._cache[sig]
        except KeyError:
            f = self.find_function(args)
            self._cache[sig] = f
            return f


    def define_function(self, f):
        for signature in self.get_func_signatures(f):
            node = self.root
            for t in signature:
                node = node.follow_type[t]

            if node.func is not None:
                code_obj = node.func[0].__code__
                raise ValueError(f"Function {f.__name__} at {code_obj.co_filename}:{code_obj.co_firstlineno} matches existing signature: {signature}!")
            node.func = f, signature


    def get_func_signatures(self, f):
        sig = inspect.signature(f)
        typesigs = []
        typesig = []
        for p in sig.parameters.values():
            # if p.kind is p.VAR_KEYWORD or p.kind is p.VAR_POSITIONAL:
            #     raise TypeError("Dispatch doesn't support *args or **kwargs yet")

            t = p.annotation
            if t is sig.empty:
                t = self.typesystem.default_type
            else:
                # Canonize to detect more collisions on construction, instead of during dispatch
                t = self.typesystem.canonize_type(t)
            # elif not isinstance(t, type):
            #     raise TypeError("Annotation isn't a type")

            if p.default is not p.empty:
                # From now on, everything is optional
                typesigs.append(list(typesig))

            typesig.append(t)

        typesigs.append(typesig)
        return typesigs

    def choose_most_specific_function(self, *funcs):
        issubclass = self.typesystem.issubclass

        class IsSubclass:
            def __init__(self, k):
                self.i, self.t = k

            def __lt__(self, other):
                return issubclass(self.t, other.t)

        most_specific_per_param = []
        sigs = [f[1] for f in funcs]
        for arg_idx, zipped_params in enumerate(zip(*sigs)):
            if all_eq(zipped_params):
                continue
            x = sorted(enumerate(zipped_params), key=IsSubclass)
            ms_i, ms_t = x[0]   # Most significant index and type
            ms_set = {ms_i}     # Init set of indexes of most significant params
            for i, t in x[1:]:
                if ms_t == t:
                    ms_set.add(i)   # Add more indexes with the same type
                elif issubclass(t, ms_t) or not issubclass(ms_t, t):
                    # Cannot resolve ordering of these two types
                    n = funcs[0][0].__name__
                    msg = f"Ambiguous dispatch in '{n}', argument #{arg_idx+1}: Unable to resolve the specificity of the types: \n\t- {t}\n\t- {ms_t}"
                    raise DispatchError(msg)

            most_specific_per_param.append(ms_set)

        # Is there only one function that matches each and every parameter?
        most_specific = set.intersection(*most_specific_per_param)
        if len(most_specific) != 1:
            n = funcs[0][0].__name__
            msg = f"Ambiguous dispatch in '{n}': Unable to resolve the specificity of the functions"
            msg += ''.join(f'\n\t- {n}{tuple(f[1])}' for f in funcs)
            raise DispatchError(msg)

        ms ,= most_specific
        return funcs[ms]

def all_eq(xs):
    a = xs[0]
    for b in xs[1:]:
        if a != b:
            return False
    return True

class TypeNode:
    def __init__(self):
        self.follow_type = defaultdict(TypeNode)
        self.func = None

    def follow_arg(self, arg, ts):
        for type_, tree in self.follow_type.items():
            if ts.isinstance(arg, type_):
                yield tree




import re
from setuptools import setup

__version__ ,= re.findall('__version__ = "(.*)"', open('runtype/__init__.py').read())

setup(
    name = "runtype",
    version = __version__,
    packages = ['runtype'],

    requires = [],
    install_requires = [],

    package_data = { '': ['*.md'] },

    test_suite = 'tests.__main__',

    # metadata for upload to PyPI
    author = "Erez Shinan",
    author_email = "erezshin@gmail.com",
    description = "Type dispatch and validation for run-time",
    license = "MIT",
    keywords = "types typing dispatch dataclass runtime",
    url = "https://github.com/erezsh/runtype",
    download_url = "https://github.com/erezsh/runtype/tarball/master",
    long_description='''
runtype is composed of several utility modules:

1. dispatch - Provides a decorator for fast multi-dispatch at run-time for functions, with sophisticated ambiguity resolution.

2. dataclass - Improves on Python's existing dataclass, by verifying the type-correctness of its attributes at run-time. Also provides a few useful methods for dataclasses.

3. isa - Provides alternative functions to `isinstance` and `issubclass`, that understand Python's `typing` module.

Runtype's integration with the `typing` module allows to use type signatures such as `List[int]`, `Optional[str]`, or `Union[int, str, Callable]`.
''',

    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 3",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Software Development :: Quality Assurance",
        "Topic :: Utilities",
        "License :: OSI Approved :: MIT License",
    ],
)

